# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines a PyTorch model to evaluate sliding puzzle boards.
"""

from typing import Optional

import torch
import torch.nn as nn

import slidingpuzzle.nn.paths as paths
import slidingpuzzle.board as board_


DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.float32
VERSION_1 = "v1"
VERSION_2 = "v2"


class Model_v1(nn.Module):
    """
    A stack of linear layers that accepts a board as input and outputs the
    estimated distance to the goal.

    Trained with::

        SGD(lr = 0.001, momentum = 0.9)
        MSELoss

    Total parameters: 1322496
    """

    def __init__(self, h: int, w: int) -> None:
        super().__init__()
        self.version = VERSION_1  # required
        self.h = h  # required
        self.w = w  # required
        hidden_size = 512
        self.input = nn.Sequential(
            nn.Flatten(),
            nn.Linear(h * w, hidden_size, dtype=DTYPE),
            nn.ReLU(),
            nn.BatchNorm1d(hidden_size, dtype=DTYPE),
        )
        n_layers = 5
        self.linears = nn.Sequential(
            *[
                nn.Sequential(
                    nn.Linear(hidden_size, hidden_size, bias=False, dtype=DTYPE),
                    nn.ReLU(),
                    nn.BatchNorm1d(hidden_size, dtype=DTYPE),
                )
                for _ in range(n_layers)
            ]
        )
        self.output = nn.Linear(hidden_size, 1, bias=False, dtype=DTYPE)

    def forward(self, x):
        x = self.input(x)
        x = self.linears(x)
        x = self.output(x)
        return x


class Model_v2(nn.Module):
    """
    A multi-headed attention-based network with about the same number of parameters
    as v1. (WIP)

    Trained with::

        SGD(lr = 0.0001, momentum = 0.9)
        MSELoss

    Total parameters: 1170767
    """

    def __init__(self, h: int, w: int) -> None:
        super().__init__()
        self.version = VERSION_2  # required
        self.h = h  # required
        self.w = w  # required
        size = h * w
        hidden_size = size * 32  # should be chosen so that is divisible by h*w
        self.flatten = nn.Flatten()
        self.embed = nn.ModuleList(
            nn.Linear(1, hidden_size, dtype=DTYPE) for _ in range(size)
        )
        self.first_attention = nn.MultiheadAttention(
            hidden_size, size, batch_first=True, dtype=DTYPE
        )
        self.first_attention_linear = nn.Linear(hidden_size, hidden_size, dtype=DTYPE)
        self.first_attention_norm = nn.BatchNorm1d(size)

        n_layers = 1
        self.attentions = nn.ModuleList(
            nn.MultiheadAttention(hidden_size, size, batch_first=True, dtype=DTYPE)
            for _ in range(n_layers)
        )
        self.linears = nn.ModuleList(
            nn.Linear(hidden_size, hidden_size, dtype=DTYPE) for _ in range(n_layers)
        )
        self.norms = nn.ModuleList(
            nn.BatchNorm1d(size, dtype=DTYPE) for _ in range(n_layers)
        )

        self.last_attention = nn.MultiheadAttention(
            hidden_size, size, batch_first=True, dtype=DTYPE
        )
        self.last_attention_linear = nn.Linear(hidden_size, 1, dtype=DTYPE)
        self.output = nn.Linear(size, 1, dtype=DTYPE)

    def forward(self, x):
        # first embed and create initial q, k, v
        x = self.flatten(x)
        x = torch.split(x, 1, dim=-1)  # break up board into individual tile tensors
        x = torch.stack([torch.relu(e(t)) for e, t in zip(self.embed, x)], dim=1)

        # attention layers
        x, _ = self.first_attention(x, x, x)
        x = torch.relu(self.first_attention_linear(x))
        x = self.first_attention_norm(x)

        for atn, lin, nrm in zip(self.attentions, self.linears, self.norms):
            x, _ = atn(x, x, x)
            x = torch.relu(lin(x))
            x = nrm(x)

        x, _ = self.last_attention(x, x, x)
        x = torch.relu(self.last_attention_linear(x))

        # output
        x = self.flatten(x)
        x = self.output(x)
        return x


def save_model(model: nn.Module, device: Optional[str] = None) -> None:
    """
    Save a frozen version of the model into the "models" dir.

    Args:
        model: The trained model.
        device: The device the model is currently loaded on. If not provied, it will
            be guessed.
    """
    if device is None:
        device = DEVICE
    model.eval()
    model.to(device)
    path = paths.get_model_path(model.h, model.w, model.version)
    board = board_.new_board(model.h, model.w)
    example_inputs = torch.tensor(board, dtype=torch.float32).unsqueeze(0).to(device)
    traced_model = torch.jit.trace(model, example_inputs)
    frozen_model = torch.jit.freeze(traced_model)
    frozen_model.save(str(path))


def load_model(
    h: int, w: int, version: str, device: Optional[str] = None
) -> torch.ScriptModule:
    """
    Reload a pre-trained frozen model.
    """
    if device is None:
        device = DEVICE
    model_path = paths.get_model_path(h, w, version)
    model = torch.jit.load(str(model_path), map_location=device)
    model.eval()
    return model


def get_num_params(model: nn.Module) -> int:
    """
    Compute the total number of parameters in a model.

    Args:
        model: A model

    Returns:
        The total number of trainable parameters
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module used to evaluate model performance vs. other heuristics.
"""
from typing import Optional

import torch
import torch.utils

from slidingpuzzle.heuristics import Heuristic
from slidingpuzzle.nn.dataset import SlidingPuzzleDataset


def accuracy(expected, predicted) -> float:
    r"""
    Helper function to estimate accuracy of sliding puzzle model outputs.
    This function replaces NaNs/infinities with 0s and uses:

    .. math::

        f(e, p) = 1 - \frac{|e - p|}{1 + |e - p|}

    It returns the sum along the batch dimension.

    Notes:
        Any NaNs/infs will be replaced by 0.

    Args:
        expected: A tensor with the expected value
        predicted: A tensor with the predicted value

    Returns:
        The accuracy as a float in the range [0, 1] summed along batch dim.
    """
    diff = torch.abs(expected - predicted)
    diff /= 1 + diff
    diff = torch.nan_to_num(diff, nan=0, posinf=0, neginf=0)
    diff = 1 - diff
    return torch.sum(diff).item()


def evaluate(
    model: torch.nn.Module, criterion, dataset: torch.utils.data.Dataset
) -> tuple[float, float]:
    """
    Runs the model on provided datatset computing average loss and accuracy.

    Args:
        model: Model to evaluate
        criterion: The criterion function to use for evaluation
        dataset: The dataset of examples

    Returns:
        A tuple(avg. loss, avg. accuracy)
    """
    model.eval()
    device = "cuda" if torch.cuda.is_available() else "cpu"
    running_loss = 0.0
    running_accuracy = 0.0
    dataloader = torch.utils.data.DataLoader(dataset)
    with torch.inference_mode():
        for batch, expected in iter(dataloader):
            predicted = model(batch.to(device))
            expected = expected.to(device)
            running_loss += criterion(predicted, expected).item()
            running_accuracy += accuracy(expected, predicted)
    return running_loss / len(dataset), running_accuracy / len(dataset)


def eval_checkpoint(
    model: torch.nn.Module,
    tag: str = "acc",
    num_iters: Optional[int] = None,
    device: Optional[str] = None,
    **kwargs,
) -> float:
    """
    Loads the provided model from the checkpoint at ``epoch`` and runs
    ``evaluate``, returning the result. If ``epoch`` is not provided, the
    latest checkpoint is used.
    """
    from slidingpuzzle.algorithms import evaluate as evaluate_
    from slidingpuzzle.nn.train import load_checkpoint
    from slidingpuzzle.nn.heuristics import set_heuristic

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    load_checkpoint(model, tag=tag)
    model.to(device)
    heuristic = set_heuristic(model)
    if num_iters is None:
        return evaluate_(model.h, model.h, weuristic=heuristic, **kwargs)
    else:
        return evaluate_(
            model.h, model.h, weuristic=heuristic, num_iters=num_iters, **kwargs
        )


def eval_heuristic(dataset: SlidingPuzzleDataset, heuristic: Heuristic) -> float:
    r"""
    Runs all examples through heuristic and evaluates accuracy as a function of
    heuristic proximity to the true distance. Can be used to compare model performance
    against other heuristics.
    """
    acc = 0.0
    for x, y in dataset:
        acc += accuracy(y, heuristic(x.numpy()))
    return acc / len(dataset)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides heuristic functions used to evaluate board states.

Each function accepts a puzzle board as input, and returns a numeric value
that indicates the estimated distance from the goal.
"""

from typing import Callable, Iterator, Optional, TypeAlias

import logging
import math

import numpy as np
import numpy.typing as npt

from slidingpuzzle.board import (
    BLANK,
    Board,
    find_blank,
    find_tile,
    get_goal_tile,
    get_goal_y,
    get_goal_x,
    get_goal_yx,
    freeze_board,
    is_solved,
    swap_tiles,
)


Heuristic: TypeAlias = Callable[[Board], int | float]

log = logging.getLogger(__name__)
manhattan_tables = {}  # (h, w): {(r, c, tile): manhattan dist}


def euclidean_distance(board: Board) -> float:
    r"""
    The distance between each tile and its destination, as measured in Euclidean space

    .. math::
        \sum_{i}^{n} \sqrt{(\text{tile}_{i, x} - \text{goal}_{i, x})^2 +
            (\text{tile}_{i, y} - \text{goal}_{i, y})^2}

    Args:
        board: The board

    Returns:
        The sum of all tile distances from their goal positions
    """
    h, w = board.shape
    dist = 0.0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            if BLANK == tile:
                continue
            goal_y, goal_x = get_goal_yx(h, w, tile)
            a, b = abs(y - goal_y), abs(x - goal_x)
            dist += math.sqrt(a**2 + b**2)
    return dist


def hamming_distance(board: Board) -> int:
    r"""
    The count of misplaced tiles.

    .. math::
        \sum_{i}^{n} \text{tile}_i \oplus \text{goal}_i

    Args:
        board: The board

    Returns:
        The number of tiles out of their goal positions
    """
    h, w = board.shape
    dist = 0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            if BLANK == tile:
                continue
            goal_y, goal_x = get_goal_yx(h, w, tile)
            if y != goal_y or x != goal_x:
                dist += 1
    return dist


def has_row_conflict(board: Board, y: int, x: int) -> bool:
    r"""
    On row ``y``, we check every tile for conflict with column ``x``.

    Args:
        board: The board:
        y: The row of interest
        x: The column we will be checking for involvement in any conflicts

    Returns:
        True if column ``x`` is involved in any conflicts on row ``y``
    """
    h, w = board.shape
    for col1 in range(w):
        tile1 = board[y, col1]
        # check conflict conditions
        if BLANK == tile1 or get_goal_y(h, w, tile1) != y:
            continue
        for col2 in range(col1 + 1, w):
            # skip tiles that don't involve our x of interest
            if col1 != x and col2 != x:
                continue
            tile2 = board[y, col2]
            # check conflict conditions
            if BLANK == tile2 or get_goal_y(h, w, tile2) != y:
                continue
            if get_goal_x(h, w, tile2) < get_goal_x(h, w, tile1):
                return True
    return False


def has_col_conflict(board: Board, y: int, x: int) -> bool:
    r"""
    See :func:`has_row_conflict`. This is identical, but swaps rows for columns.
    """
    h, w = board.shape
    for row1 in range(h):
        tile1 = board[row1, x]
        # check conflict conditions
        if BLANK == tile1 or get_goal_x(h, w, tile1) != x:
            continue
        for row2 in range(row1 + 1, h):
            if row1 != y and row2 != y:
                continue
            tile2 = board[row2, x]
            # check conflict conditions
            if BLANK == tile2 or get_goal_x(h, w, tile2) != x:
                continue
            if get_goal_y(h, w, tile2) < get_goal_y(h, w, tile1):
                return True
    return False


def has_conflict(board: Board, y: int, x: int):
    r"""
    Combines :func:`has_row_conflict` and :func:`has_col_conflict` for convenience.
    """
    return has_row_conflict(board, y, x) or has_col_conflict(board, y, x)


def last_moves_distance(board: Board, relaxed: bool = True) -> int:
    r"""
    Immediately before the last move, one of the tiles adjacent to the blank tile must
    be in the corner. If not, we must make at least 2 more moves to shuffle.
    This is similar to :func:`corner_tiles_distance`, but specifically targets the goal
    corner, which has slightly different constraints.This function also ensures
    admissibility with :func:`manhattan_distance` and :func:`linear_conflict_distance`.
    If you aren't using this with other heuristics, you can set ``relaxed=True``.

    Note:
        Currently this only considers the very last moves, not the moves also prior to
        the last moves (2-removed) as discussed in Korf.

    Args:
        board: The board
        relaxed: If False, can be safely combined with  :func:`manhattan_distance` and
            :func:`linear_conflict_distance`.

    Returns:
        2 if neither of the final tiles to move are in the final corner, else 0.
    """
    h, w = board.shape
    adj1 = get_goal_tile(h, w, (h - 2, w - 1))  # tile in row next to goal
    adj2 = get_goal_tile(h, w, (h - 1, w - 2))  # tile in col next to goal
    corner = board[-1, -1]
    if corner == adj1 or corner == adj2 or is_solved(board):
        return 0
    if relaxed:
        return 2

    # coords of these tiles on cur. board
    adj1_y, adj1_x = find_tile(board, adj1)
    adj2_y, adj2_x = find_tile(board, adj2)
    # prevent interaction with Manhattan
    if adj1_y == h - 1 or adj2_x == w - 1:
        return 0
    # prevent interaction with linear conflict distance
    if adj1_y == h - 2 and has_row_conflict(board, adj1_y, adj1_x):
        return 0
    if adj2_x == w - 2 and has_col_conflict(board, adj2_y, adj2_x):
        return 0

    # if the tiles that must make the final move are not located next to their goal
    # row/col, they will need to shuffle before the final move
    return 2


def corner_tiles_distance(board: Board, relaxed: bool = True) -> int:
    r"""
    If tiles that belong in corners are not there, but the adjacent tiles are correct,
    additional reshuffling will need to occur. We must ensure that row/col conflicts
    have not already accounted for this in order to combine admissibily with the
    :func:`linear_conflict_distance`.

    Note:
        This heuristic is inadmissible if used on boards of insufficient size. For
        example, on a 3x3 board, the corners share some adjacent tiles and this
        will result in overcounting.

    Args:
        board: The board
        relaxed: If False, relevant tiles will be checked for linear conflicts.

    Returns:
        The additional distance required to shuffle corners into position.
    """
    h, w = board.shape
    dist = 0

    def check_adjacent(y, x) -> int:
        r""" "
        Helper to ensure this corner-adjacent tile is in its correct position
        and is not involved in a conflict.
        """
        adj = get_goal_tile(h, w, (y, x))
        if adj != board[y, x] or (not relaxed and has_conflict(board, y, x)):
            return 0
        return 2

    def check_corner(y, x, adj1, adj2) -> int:
        r"""
        Checks if:
            - Corner tile is out of position
            - Corner adjacents are in correct position

        Returns:
            Either 0, 2, or 4 to be added to distance.
        """
        corner = get_goal_tile(h, w, (y, x))
        if BLANK != board[y, x] and corner != board[y, x]:
            return check_adjacent(*adj1) + check_adjacent(*adj2)
        return 0

    # top-left corner
    dist += check_corner(0, 0, (0, 1), (1, 0))
    # top-right corner
    dist += check_corner(0, w - 1, (0, w - 2), (1, w - 1))
    # bottom-left corner
    dist += check_corner(h - 1, 0, (h - 1, 1), (h - 2, 0))

    return dist


def linear_conflict_distance(board: Board, optimized: bool = True) -> int:
    r"""
    Starts with Manhattan distance, then for each row and column, the number of tiles
    "in conflict" are identified, and 2 * this number is added to the total distance.
    (It will take at least 2 additional moves to reshuffle the conflicting tiles into
    their correct positions.) This is an admissible improvement over
    :func:`manhattan_distance` (`Hansson, Mayer, Young, 1985 <hannson_>`_).
    The :func:`last_moves_distance` and :func:`corner_tiles_distance` are included
    (`Korf and Taylor, 1996 <korf_>`_).

    Args:
        board: The board
        optimized: If False, will not include :func:`manhattan_distance`,
            :func:`last_moves_distance` and :func:`corner_tiles_distance`. It will
            return only the base number of linear conflicts.

    Returns:
        Estimated distance to goal.

    .. _hannson:
        https://academiccommons.columbia.edu/doi/10.7916/D8154QZT/download

    .. _korf:
        https://www.aaai.org/Library/AAAI/1996/aaai96-178.php
    """
    h, w = board.shape
    dist = manhattan_distance(board) if optimized else 0

    def line_generator() -> Iterator[tuple[npt.NDArray, npt.NDArray]]:
        r"""
        Helper to generate all lines on a board (rows followed by columns) along with
        a bool specifying whether the tile is in its goal line.

        Returns:
            A tuple of line index, line values, and goal positions.
        """
        for y, row in enumerate(board):
            yield row, np.fromiter(
                (y == get_goal_y(h, w, tile) for tile in row), dtype=bool
            )
        for x, col in enumerate(board.T):
            yield col, np.fromiter(
                (x == get_goal_x(h, w, tile) for tile in col), dtype=bool
            )

    def get_line_conflicts(line, goals) -> npt.NDArray[np.integer]:
        r"""
        Helper to return a list of conflict counts for a line.

        Args:
            line: The tile values in the line
            goals: True/False if goal position for each tile is in this line

        Returns:
            Conflict counts for each tile
        """
        conflicts = np.zeros(len(line), dtype=int)
        for pos1, (tile1, goal1) in enumerate(zip(line, goals)):
            # check if this tile is in it's goal line
            if not goal1 or BLANK == tile1:
                continue
            # now check tiles after this
            next_pos = pos1 + 1
            for pos2, (tile2, goal2) in enumerate(
                zip(line[next_pos:], goals[next_pos:]), next_pos
            ):
                # same checks for tile2
                if not goal2 or BLANK == tile2:
                    continue
                # check if these tiles are in conflict
                if tile2 < tile1:
                    conflicts[pos1] += 1
                    conflicts[pos2] += 1
        return conflicts

    for line, goals in line_generator():
        line = np.copy(line)  # don't modify original board
        while np.any(conflicts := get_line_conflicts(line, goals)):
            dist += 2
            line[np.argmax(conflicts)] = BLANK  # remove largest conflict

    # we can't use corner distance on tiny boards, b/c it not only conflicts with
    # last moves distance, it also conflicts with itself
    if optimized:
        if h > 3 and w > 3:
            dist += corner_tiles_distance(board, relaxed=False)
        dist += last_moves_distance(board, relaxed=False)

    return dist


def prepare_manhattan_table(h, w) -> dict[tuple[int, int, int], int]:
    log.debug(f"Preparing first use of Manhattan table {h}x{w}x{h*w}...")
    table = {}
    for y in range(h):
        for x in range(w):
            for tile in range(h * w):
                if BLANK == tile:
                    table[(y, x, tile)] = 0
                else:
                    goal_y, goal_x = get_goal_yx(h, w, tile)
                    table[(y, x, tile)] = abs(y - goal_y) + abs(x - goal_x)
    return table


def manhattan_distance(board: Board) -> int:
    r"""
    The minimum number of moves needed to restore the board to the goal state, if tiles
    could be moved through each other.

    .. math::
        \sum_{i}^{n} |\text{tile}_{i, x} - \text{goal}_{i, x}| +
            |\text{tile}_{i, y} - \text{goal}_{i, y}|

    Args:
        board: The board

    Returns:
        The sum of all tile distances from their goal positions
    """
    h, w = board.shape
    table = manhattan_tables.get((h, w), None)
    if table is None:
        table = prepare_manhattan_table(h, w)
        manhattan_tables[(h, w)] = table

    dist = 0
    for y, row in enumerate(board):
        for x, tile in enumerate(row):
            dist += table[(y, x, tile)]
    return dist


def random_distance(board: Board) -> int:
    r"""
    A random distance computed as a hash of the board state. Useful as a baseline.

    .. math::
        hash(\text{board})

    Args:
        board: The board

    Returns:
        A random distance that is consistent for a given board state
    """
    return hash(freeze_board(board))


def relaxed_adjacency_distance(board: Board) -> int:
    r"""
    Repeatedly swap misplaced tiles into their goal positions, ignoring legal board
    distance and rules. We repeat this process until the board is solved. This
    heuristic is a slight improvement to Hamming distance.

    Args:
        board: The board

    Returns:
        The estimated distance to goal.
    """
    h, w = board.shape
    board = np.copy(board)
    dist = 0

    def swap_first_misplaced(blank_yx: tuple[int, int]) -> None:
        for y in range(h):
            for x in range(w):
                if get_goal_yx(h, w, board[y][x]) != (y, x):
                    swap_tiles(board, (y, x), blank_yx)

    while not is_solved(board):
        blank_yx = find_blank(board)
        if blank_yx == get_goal_yx(h, w, BLANK):
            swap_first_misplaced(blank_yx)
        else:
            swap_tiles(board, get_goal_tile(h, w, blank_yx), blank_yx)
        dist += 1

    return dist



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for creating, saving, and loading board datasets.
"""

from typing import Iterable, Optional, TypeAlias

import collections
import gc
import json
import logging

import torch
import torch.utils
import tqdm

import slidingpuzzle.algorithms as algorithms
import slidingpuzzle.nn.paths as paths
from slidingpuzzle.board import (
    Board,
    FrozenBoard,
    find_blank,
    freeze_board,
    new_board,
    shuffle,
    solution_as_tiles,
    swap_tiles,
    visit,
)
from slidingpuzzle.state import State


Example: TypeAlias = tuple[FrozenBoard, list[int]]
log = logging.getLogger(__name__)


class SlidingPuzzleDataset(torch.utils.data.Dataset):
    def __init__(self, examples: Iterable[Example]) -> None:
        super().__init__()
        self.examples = [
            (
                torch.tensor(x, dtype=torch.float32),
                torch.tensor([len(y)], dtype=torch.float32),
            )
            for x, y in examples
        ]

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, idx):
        return self.examples[idx]


def all_examples(
    h: int, w: int, start: int = 0, stop: Optional[int] = None
) -> list[Example]:
    """
    Helper to generate all training examples for a board size by performing reverse BFS.

    .. note::
        It is safe to interrupt this function any time with a keyboard interrupt and
        receive back the examples generated so far.

    Args:
        h: Height of the board
        w: Width of the board
        start: Start index (0 starts at the initial solved board)
        stop: Index to stop at (``None`` will product all boards)

    Returns:
         A list of examples.
    """
    visited: set[FrozenBoard] = set()
    examples = []
    board = new_board(h, w)
    initial_state = State(board, find_blank(board))
    unvisited = collections.deque([initial_state])
    count = 0
    pbar = tqdm.tqdm()
    try:
        while unvisited and (stop is None or count < stop):
            count += 1
            state = unvisited.popleft()
            if count < start or visit(visited, state.board):
                continue
            board = freeze_board(state.board)
            solution = tuple(solution_as_tiles(state.board, reversed(state.history)))
            examples.append((board, solution))
            pbar.update(1)
            next_states = algorithms.get_next_states(state)
            # patch the history to reflect that this search is backwards
            for ns in next_states:
                ns.history[-1] = state.blank_pos
            unvisited.extend(next_states)
    except KeyboardInterrupt:
        log.info(f"Example generation interrupted, returning {len(examples)} examples.")
    finally:
        pbar.close()
    return examples


def random_examples(
    h: int,
    w: int,
    num_examples: int,
    ignore_examples: Optional[list[Example]] = None,
    **kwargs,
) -> list[Example]:
    """
    Constructs a list of training examples, which are tuples of:
        (board, solution)

    Args:
        h: Height of board.
        w: Width of board.
        num_examples: Number of examples to produce.
        ignore_examples: If any example produced matches one in ``ignore_examples``,
            it will be discarded.
        kwargs: Args to pass to the search algorithm used to find board solutions.

    Returns:
        The list of training examples.
    """
    visited: set[FrozenBoard] = set()
    examples = []
    if ignore_examples is not None:
        dupe_found = False
        for board, _ in ignore_examples:
            if visit(visited, board):
                dupe_found = True
        if dupe_found:
            log.warning("Duplicate found in prior examples.")

    def add_examples(board: Board) -> None:
        # free memory from prior searches
        gc.collect()

        # find a path to use as an accurate training reference
        result = algorithms.search(board, **kwargs)

        # we can use all intermediate boards as examples
        while len(examples) < num_examples:
            solution = tuple(solution_as_tiles(result.board, result.solution))
            examples.append((freeze_board(board), solution))
            pbar.update(1)
            if not len(result.solution):
                break
            swap_tiles(board, result.solution.pop(0))
            if visit(visited, board):
                break

    try:
        with tqdm.tqdm(total=num_examples) as pbar:
            while len(examples) < num_examples:
                board = shuffle(new_board(h, w))
                if not visit(visited, board):
                    add_examples(board)
    except KeyboardInterrupt:
        log.info(f"Example generation interrupted, returning {len(examples)} examples.")

    return examples


def load_examples(h: int, w: int, examples_file: Optional[str] = None) -> list[Example]:
    """
    Loads examples from a JSON file.
    """
    if examples_file is None:
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "rt") as fp:
        return json.load(fp)


def save_examples(examples: list[Example], examples_file: Optional[str] = None) -> None:
    """
    Save a list of examples to disk as JSON.
    """
    if 0 == len(examples):
        return
    if examples_file is None:
        board, _ = examples[0]
        h, w = len(board), len(board[0])
        examples_file = paths.get_examples_path(h, w)
    with open(examples_file, "wt") as fp:
        json.dump(examples, fp)


def get_examples(
    h: int, w: int, num_examples: int, prior_examples: list[Example], **kwargs
) -> list[Example]:
    """
    Returns ``num_examples`` total unique examples, starting with ``prior_examples`` if
    they are provided. May truncate or produce new examples as necessary.
    """
    examples = list(prior_examples)
    if len(examples) > num_examples:
        examples = examples[:num_examples]
    if len(examples) < num_examples:
        num_new_examples = num_examples - len(prior_examples)
        log.info(f"Constructing {num_new_examples} new examples...")
        examples.extend(
            random_examples(h, w, num_new_examples, prior_examples, **kwargs)
        )
    return examples


def build_or_load_dataset(
    h: int,
    w: int,
    num_examples: Optional[int] = None,
    examples_file: Optional[str] = None,
    **kwargs,
) -> SlidingPuzzleDataset:
    """
    Loads examples, constructs a SlidingPuzzleDataset from them, and returns it.
    If there is a mismatch between the requested num_examples and the loaded
    examples file, examples may be truncated or new examples may be constructed and
    saved to disk. No duplicate examples are produced.

    Args:
        h: The height of the board to locate a dataset for
        w: The width of the board to locate a dataset for
        num_examples: The total number of examples desired. If there are too many,
            examples will be truncated. If there are too few, new examples will be
            constructed. If ``None``, the entire dataset loaded will be used.
        examples_file: The name of the examples file to save or load.
        kwargs: Will be forwaded to :func:`make_examples` if it is called.

    Returns:
        A dataset for the requested puzzle size.
    """
    log.info("Loading dataset...")
    try:
        examples = load_examples(h, w, examples_file)
        log.info(f"Dataset loaded with {len(examples)} examples.")
    except FileNotFoundError:
        log.info("No dataset found.")
        examples = []

    # if a specific number was not requested, use all loaded examples
    if num_examples is None:
        num_examples = len(examples)

    # if we were asked for a different number of examples than we have on hand
    if len(examples) != num_examples:
        new_examples = get_examples(h, w, num_examples, examples, **kwargs)
        # if we made new examples, save them for next time
        if len(new_examples) > len(examples):
            print(new_examples)
            save_examples(new_examples, examples_file)
            log.info("Dataset saved.")
        examples = new_examples

    return SlidingPuzzleDataset(examples)


def load_dataset(
    h: int, w: int, examples_file: Optional[str] = None
) -> SlidingPuzzleDataset:
    """
    Loads examples, constructs a SlidingPuzzleDataset from them, and returns it.

    Args:
        h: The height of the board to locate a dataset for
        w: The width of the board to locate a dataset for
        examples_file: Path to file. Default will check local ``datasets`` directory.

    Returns:
        A dataset for the requested puzzle size.
    """
    log.info("Loading dataset...")
    try:
        examples = load_examples(h, w, examples_file)
        log.info(f"Dataset loaded with {len(examples)} examples.")
    except FileNotFoundError:
        log.info("No dataset found.")
        examples = []
    return SlidingPuzzleDataset(examples)



# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import pytest

from slidingpuzzle import *


def test_corner_tiles_distance():
    board = new_board(3, 3)
    assert corner_tiles_distance(board) == 0

    # top-left corner
    board = from_rows([8, 5, 3], [4, 2, 6], [7, 1, 0])
    assert corner_tiles_distance(board) == 2

    board = from_rows([8, 2, 3], [4, 5, 6], [7, 0, 1])
    assert corner_tiles_distance(board) == 4

    # top-right corner
    board = from_rows([1, 2, 8], [4, 6, 5], [7, 0, 3])
    assert corner_tiles_distance(board) == 2

    board = from_rows([1, 2, 8], [3, 5, 6], [4, 0, 7])
    assert corner_tiles_distance(board) == 4

    # bottom-left corner
    board = from_rows([1, 2, 3], [4, 7, 6], [5, 0, 8])
    assert corner_tiles_distance(board) == 2

    board = from_rows([1, 2, 3], [4, 7, 6], [5, 8, 0])
    assert corner_tiles_distance(board) == 4


def test_euclidean_distance():
    board = new_board(3, 5)
    assert euclidean_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    c = math.sqrt((2 * 1) ** 2 + (2 * 2) ** 2)
    assert euclidean_distance(board) == c


def test_hamming_distance():
    board = new_board(5, 3)
    assert hamming_distance(board) == 0

    swap_tiles(board, (0, 1), (0, 0))
    assert hamming_distance(board) == 2


def test_last_moves_distance():
    board = from_rows([1, 2, 3], [4, 5, 6], [7, 0, 8])
    assert last_moves_distance(board) == 0

    board = from_rows([8, 6, 0], [4, 5, 2], [7, 3, 1])
    assert last_moves_distance(board) == 2


def test_linear_conflict_distance():
    board = from_rows([1, 2, 3], [4, 5, 6], [7, 0, 8])
    assert linear_conflict_distance(board, optimized=False) == 0
    assert linear_conflict_distance(board) == 1

    board = from_rows([2, 1, 3], [4, 5, 6], [7, 8, 0])
    assert linear_conflict_distance(board, optimized=False) == 2
    assert linear_conflict_distance(board) == 6

    board = from_rows([4, 2, 3], [1, 5, 6], [7, 8, 0])
    assert linear_conflict_distance(board, optimized=False) == 2
    assert linear_conflict_distance(board) == 6

    board = from_rows([2, 1, 3], [4, 5, 6], [7, 0, 8])
    assert linear_conflict_distance(board, optimized=False) == 2
    assert linear_conflict_distance(board) == 5

    board = from_rows([1, 2, 3], [6, 5, 4], [7, 8, 0])
    assert linear_conflict_distance(board, optimized=False) == 4
    assert linear_conflict_distance(board) == 8


def test_manhattan_distance():
    board = new_board(3, 5)
    assert manhattan_distance(board) == 0

    swap_tiles(board, (1, 2), (0, 0))
    assert manhattan_distance(board) == 6


def test_relaxed_adjacency_distance():
    h, w = 3, 3
    b = new_board(h, w)
    swap_tiles(b, (h - 1, w - 1), (0, 0))
    assert relaxed_adjacency_distance(b) == 1

    b = new_board(h, w)
    swap_tiles(b, (0, 0), (0, 1))
    swap_tiles(b, (0, 0), (h - 1, w - 1))
    swap_tiles(b, (0, 0), (1, 0))
    assert relaxed_adjacency_distance(b) == 3


@pytest.mark.slow
def test_heuristic_behavior():
    # we compute avg generated nodes over multiple runs to confirm that
    # heuristic behavior is in line with expectations
    generated_avg = compare(
        3,
        3,
        ha=linear_conflict_distance,
        hb=manhattan_distance,
    )
    assert generated_avg[0] < generated_avg[1]

    generated_avg = compare(
        3, 3, num_iters=4, ha=manhattan_distance, hb=hamming_distance
    )
    assert generated_avg[0] < generated_avg[1]

    # lcd/manhattan/euclidean are good contenders, so we don't compare them
    generated_avg = compare(
        3, 3, num_iters=4, ha=euclidean_distance, hb=hamming_distance
    )
    assert generated_avg[0] < generated_avg[1]


@pytest.mark.slow
@pytest.mark.parametrize("hw", [(3, 3)])
def test_heuristic_admissibility(hw):
    h, w = hw
    # validate that solutions are in line with BFS
    # this does not guarantee admissibility, it's just an empirical sanity check
    boards = [shuffle(new_board(h, w)) for _ in range(50)]
    optimal = [len(search(b, "bfs").solution) for b in boards]
    for h in (linear_conflict_distance, manhattan_distance):
        for b, o in zip(boards, optimal):
            assert len(search(b, heuristic=h).solution) == o, b


@pytest.mark.skip
@pytest.mark.slow
@pytest.mark.parametrize("hw", [(3, 3)])
def test_linear_conflict_distance_exhaustive(hw):
    h, w = hw
    start, stop = 0, None
    for i, b in enumerate(board_generator(h, w, start, stop)):
        expected = len(search(b, heuristic=manhattan_distance).solution)
        actual = linear_conflict_distance(b)
        assert expected >= actual, f"board #{start + i}: {b}"


@pytest.mark.skip
@pytest.mark.slow
@pytest.mark.parametrize("hw", [(3, 3), (4, 4)])
def test_linear_conflict_distance_exhaustive_dataset(hw):
    r"""
    This test is ths same as above, but faster if a dataset file is available
    instead of re-generating and solving all boards from scratch.
    """
    import json

    h, w = hw
    with open(f"datasets/examples_{h}x{w}.json") as fp:
        db = json.load(fp)
    for i, (board, solution) in enumerate(db):
        board = from_rows(*board)
        expected = len(solution)
        actual = linear_conflict_distance(board)
        assert expected >= actual, f"board #{i}: {board}, {solution}"



from slidingpuzzle import *
#b=from_rows([0,1,5],[4,6,7],[2,8,3])
b=from_rows([2, 1, 3], [4, 5, 6], [7, 8, 0])
s1 = search(b, heuristic=linear_conflict_distance)
s2 = search(b, heuristic=manhattan_distance)
print("true distance:", len(s2.solution))
print("our distance:", len(s1.solution))


# Copyright 2022 Stephen Dunn

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for dealing with paths required during model training and inference.
"""

import pathlib
import shutil


CHECKPOINT_DIR = "checkpoints"
DATASET_DIR = "datasets"
MODELS_DIR = "models"
TENSORBOARD_DIR = "tensorboard"


def clear_training(h: int, w: int) -> None:
    r"""
    Removes checkpoints and tensorboard logs.
    """
    shutil.rmtree(get_checkpoint_dir(h, w), ignore_errors=True)
    shutil.rmtree(get_log_dir(h, w), ignore_errors=True)


def get_board_size_str(h: int, w: int) -> str:
    """
    Helper to get a string encoding height and width.
    """
    return f"{h}x{w}"


def get_path(dirname: str, filename: str) -> pathlib.Path:
    """
    Creates intermediate directorys for dirname and returns the full path from dirname
    to filename.

    Args:
        dirname: A directory or path to a directory.
        filename: The filename the path will point to.

    Returns:
        The path object.
    """
    dirpath = pathlib.Path(dirname)
    dirpath.mkdir(exist_ok=True, parents=True)
    return dirpath / filename


def get_checkpoint_dir(h: int, w: int) -> pathlib.Path:
    """
    Get the path to store checkpoints for this board size.

    Args:
        h: Board height
        w: Board width

    Returns:
        The checkpoint dir path
    """
    return pathlib.Path(CHECKPOINT_DIR) / get_board_size_str(h, w)


def get_checkpoint_path(h: int, w: int, tag: str) -> pathlib.Path:
    """
    Get the path to a checkpoint, given a board size and optional tag.

    Args:
        h: Board height
        w: Board width
        tag: Checkpoint tag to load

    Returns:
        The path to the checkpoint file
    """
    return get_path(get_checkpoint_dir(h, w), f"checkpoint_{tag}")


def get_examples_path(h: int, w: int) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(DATASET_DIR, f"examples_{board_size_str}.json")


def get_model_path(h: int, w: int, version: str) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(MODELS_DIR, f"{version}_{board_size_str}.pt")


def get_log_dir(h: int, w: int) -> pathlib.Path:
    board_size_str = get_board_size_str(h, w)
    return get_path(TENSORBOARD_DIR, f"slidingpuzzle_{board_size_str}")
